# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import itertools
from pathlib import Path

import cv2
import numpy as np
import PIL.Image
import torch.nn as nn
import torch
import torchvision
from torchvision.models import resnet50
import torchvision.transforms as tr

import nevergrad as ng
import nevergrad.common.typing as tp
from nevergrad.common import errors
from .. import base
from . import imagelosses

# pylint: disable=abstract-method,too-many-instance-attributes,too-many-arguments


class Image(base.ExperimentFunction):
    def __init__(
        self,
        problem_name: str = "recovering",
        index: int = 0,
        loss: tp.Type[imagelosses.ImageLoss] = imagelosses.SumAbsoluteDifferences,
        with_pgan: bool = False,
        num_images: int = 1,
    ) -> None:
        """
        problem_name: the type of problem we are working on.
           recovering: we directly try to recover the target image.§
        index: the index of the problem, inside the problem type.
           For example, if problem_name is "recovering" and index == 0,
           we try to recover the face of O. Teytaud.
        """

        # Storing high level information.
        self.domain_shape = (226, 226, 3)
        self.problem_name = problem_name
        self.index = index
        self.with_pgan = with_pgan
        self.num_images = num_images

        # Storing data necessary for the problem at hand.
        assert problem_name == "recovering"  # For the moment we have only this one.
        assert index == 0  # For the moment only 1 target.
        # path = os.path.dirname(__file__) + "/headrgb_olivier.png"
        path = Path(__file__).with_name("headrgb_olivier.png")
        image = PIL.Image.open(path).resize((self.domain_shape[0], self.domain_shape[1]), PIL.Image.BICUBIC)
        self.data = np.asarray(image)[:, :, :3]  # 4th Channel is pointless here, only 255.
        # parametrization
        if not with_pgan:
            assert num_images == 1
            array = ng.p.Array(init=128 * np.ones(self.domain_shape), mutable_sigma=True)
            array.set_mutation(sigma=35)
            array.set_bounds(lower=0, upper=255.99, method="clipping", full_range_sampling=True)
            max_size = ng.p.Scalar(lower=1, upper=200).set_integer_casting()
            array = ng.ops.mutations.Crossover(axis=(0, 1), max_size=max_size)(array).set_name(
                ""
            )  # type: ignore
            super().__init__(loss(reference=self.data), array)
        else:
            self.pgan_model = torch.hub.load(
                "facebookresearch/pytorch_GAN_zoo:hub",
                "PGAN",
                model_name="celebAHQ-512",
                pretrained=True,
                useGPU=False,
            )
            self.domain_shape = (num_images, 512)  # type: ignore
            initial_noise = np.random.normal(size=self.domain_shape)
            self.initial = np.random.normal(size=(1, 512))
            self.target = np.random.normal(size=(1, 512))
            array = ng.p.Array(init=initial_noise, mutable_sigma=True)
            array.set_mutation(sigma=35.0)
            array = ng.ops.mutations.Crossover(axis=(0, 1))(array).set_name("")
            self._descriptors.pop("use_gpu", None)
            super().__init__(self._loss_with_pgan, array)

        assert self.multiobjective_upper_bounds is None
        self.add_descriptors(loss=loss.__class__.__name__)
        self.loss_function = loss(reference=self.data)

    def _generate_images(self, x: np.ndarray) -> np.ndarray:
        """Generates images tensor of shape [nb_images, x, y, 3] with pixels between 0 and 255"""
        # pylint: disable=not-callable
        noise = torch.tensor(x.astype("float32"))
        return (
            ((self.pgan_model.test(noise).clamp(min=-1, max=1) + 1) * 255.99 / 2)
            .permute(0, 2, 3, 1)
            .cpu()
            .numpy()[:, :, :, [2, 1, 0]]
        )

    def interpolate(self, base_image: np.ndarray, target: np.ndarray, k: int, num_images: int) -> np.ndarray:
        if num_images == 1:
            return target
        coef1 = k / (num_images - 1)
        coef2 = (num_images - 1 - k) / (num_images - 1)
        return coef1 * base_image + coef2 * target

    def _loss_with_pgan(self, x: np.ndarray, export_string: str = "") -> float:
        loss = 0.0
        factor = 1 if self.num_images < 2 else 10  # Number of intermediate images.
        num_total_images = factor * self.num_images
        for i in range(num_total_images):
            base_i = i // factor
            # We generate num_images images. The last one is close to target, the first one is close to initial if num_images > 1.
            base_image = self.interpolate(self.initial, self.target, i, num_total_images)
            movability = 0.5  # If only one image, then we move by 0.5.
            if self.num_images > 1:
                movability = 4 * (
                    0.25 - (i / (num_total_images - 1) - 0.5) ** 2
                )  # 1 if i == num_total_images/2, 0 if 0 or num_images-1
            moving = (
                movability
                * np.sqrt(self.dimension)
                * np.expand_dims(x[base_i], 0)
                / (1e-10 + np.linalg.norm(x[base_i]))
            )
            base_image = moving if self.num_images == 1 else base_image + moving
            image = self._generate_images(base_image).squeeze(0)
            image = cv2.resize(image, dsize=(226, 226), interpolation=cv2.INTER_NEAREST)
            if export_string:
                cv2.imwrite(f"{export_string}_image{i}_{num_total_images}_{self.num_images}.jpg", image)
            assert image.shape == (226, 226, 3), f"{x.shape} != {(226, 226, 3)}"
            loss += self.loss_function(image)
        return loss

    def export_to_images(self, x: np.ndarray, export_string: str = "export"):
        self._loss_with_pgan(x, export_string=export_string)


# #### Adversarial attacks ##### #


class Normalize(nn.Module):
    def __init__(self, mean: tp.ArrayLike, std: tp.ArrayLike) -> None:
        super().__init__()
        self.mean = torch.Tensor(mean)
        self.std = torch.Tensor(std)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return (x - self.mean.type_as(x)[None, :, None, None]) / self.std.type_as(x)[None, :, None, None]


class Resnet50(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.norm = Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
        self.model = resnet50(pretrained=True)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.model(self.norm(x))


class TestClassifier(nn.Module):
    def __init__(self, image_size: int = 224) -> None:
        super().__init__()
        self.model = nn.Linear(image_size * image_size * 3, 10)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.model(x.view(x.shape[0], -1))


# pylint: disable=too-many-arguments,too-many-instance-attributes
class ImageAdversarial(base.ExperimentFunction):
    def __init__(
        self,
        classifier: nn.Module,
        image: torch.Tensor,
        label: int = 0,
        targeted: bool = False,
        epsilon: float = 0.05,
    ) -> None:
        # TODO add crossover params in args + criterion
        """
        params : needs to be detailed
        """
        self.targeted = targeted
        self.epsilon = epsilon
        self.image = image  # if (image is not None) else torch.rand((3, 224, 224))
        self.label = torch.Tensor([label])  # if (label is not None) else torch.Tensor([0])
        self.label = self.label.long()
        self.classifier = classifier  # if (classifier is not None) else Classifier()
        self.classifier.eval()
        self.criterion = nn.CrossEntropyLoss()
        self.imsize = self.image.shape[1]

        array = ng.p.Array(
            init=np.zeros(self.image.shape),
            mutable_sigma=True,
        ).set_name("")
        array.set_mutation(sigma=self.epsilon / 10)
        array.set_bounds(lower=-self.epsilon, upper=self.epsilon, method="clipping", full_range_sampling=True)
        max_size = ng.p.Scalar(lower=1, upper=200).set_integer_casting()
        array = ng.p.mutation.Crossover(axis=(1, 2), max_size=max_size)(array)
        super().__init__(self._loss, array)
        self._descriptors.pop("label")  # automatically added by __new__, but not needed

    def _loss(self, x: np.ndarray) -> float:
        output_adv = self._get_classifier_output(x)
        value = float(self.criterion(output_adv, self.label).item())
        return value * (1.0 if self.targeted else -1.0)

    def _get_classifier_output(self, x: np.ndarray) -> tp.Any:
        # call to the classifier given the input array
        y = torch.Tensor(x)
        image_adv = torch.clamp(self.image + y, 0, 1)
        image_adv = image_adv.view(1, 3, self.imsize, self.imsize)
        return self.classifier(image_adv)

    def evaluation_function(self, *recommendations: ng.p.Parameter) -> float:
        """Returns wether the attack worked or not"""
        assert len(recommendations) == 1, "Should not be a pareto set for a singleobjective function"
        x = recommendations[0].value
        output_adv = self._get_classifier_output(x)
        _, pred = torch.max(output_adv, axis=1)
        actual = int(self.label)
        return float(pred != actual if self.targeted else pred == actual)

    @classmethod
    def make_folder_functions(
        cls,
        folder: tp.Optional[tp.PathLike],
        model: str = "resnet50",
    ) -> tp.Generator["ImageAdversarial", None, None]:
        """

        Parameters
        ----------
        folder: str or None
            folder to use for reference images. If None, 1 random image is created.
        model: str
            model name to use

        Yields
        ------
        ExperimentFunction
            an experiment function corresponding to 1 of the image of the provided folder dataset.
        """
        assert model in {"resnet50", "test"}
        tags = {"folder": "#FAKE#" if folder is None else Path(folder).name, "model": model}
        classifier: tp.Any = Resnet50() if model == "resnet50" else TestClassifier()
        classifier.eval()

        transform = tr.Compose([tr.Resize(256), tr.CenterCrop(224), tr.ToTensor()])
        if folder is None:
            x = torch.zeros(1, 3, 224, 224)
            _, pred = torch.max(classifier(x), axis=1)
            data_loader: tp.Iterable[tp.Tuple[tp.Any, tp.Any]] = [(x, pred)]
        elif Path(folder).is_dir():
            ifolder = torchvision.datasets.ImageFolder(folder, transform=transform)
            data_loader = torch.utils.data.DataLoader(
                ifolder, batch_size=1, shuffle=True, num_workers=8, pin_memory=True
            )
        else:
            raise ValueError(f"{folder} is not a valid folder.")

        for data, target in itertools.islice(data_loader, 0, 100):
            _, pred = torch.max(classifier(data), axis=1)
            if pred == target:
                func = cls(
                    classifier=classifier, image=data[0], label=int(target), targeted=False, epsilon=0.05
                )
                func.add_descriptors(**tags)
                yield func


class ImageFromPGAN(base.ExperimentFunction):
    """
    Creates face images using a GAN from pytorch GAN zoo trained on celebAHQ and optimizes the noise vector of the GAN

    Parameters
    ----------
    problem_name: str
        the type of problem we are working on.
    initial_noise: np.ndarray
        the initial noise of the GAN. It should be of dimension (1, 512). If None, it is defined randomly.
    use_gpu: bool
        whether to use gpus to compute the images
    loss: ImageLoss
        which loss to use for the images (default: Koncept512)
    mutable_sigma: bool
        whether the sigma should be mutable
    sigma: float
        standard deviation of the initial mutations
    """

    def __init__(
        self,
        initial_noise: tp.Optional[np.ndarray] = None,
        use_gpu: bool = False,
        loss: tp.Optional[imagelosses.ImageLoss] = None,
        mutable_sigma: bool = True,
        sigma: float = 35,
    ) -> None:
        if loss is None:
            loss = imagelosses.Koncept512()
        if not torch.cuda.is_available():
            use_gpu = False
        # Storing high level information..
        if os.environ.get("CIRCLECI", False):
            raise errors.UnsupportedExperiment("ImageFromPGAN is not well supported in CircleCI")
        self.pgan_model = torch.hub.load(
            "facebookresearch/pytorch_GAN_zoo:hub",
            "PGAN",
            model_name="celebAHQ-512",
            pretrained=True,
            useGPU=use_gpu,
        )

        self.domain_shape = (1, 512)
        if initial_noise is None:
            initial_noise = np.random.normal(size=self.domain_shape)
        assert initial_noise.shape == self.domain_shape, (
            f"The shape of the initial noise vector was {initial_noise.shape}, "
            f"it should be {self.domain_shape}"
        )

        array = ng.p.Array(init=initial_noise, mutable_sigma=mutable_sigma)
        # parametrization
        array.set_mutation(sigma=sigma)
        array = ng.ops.mutations.Crossover(axis=(0, 1))(array).set_name("")

        super().__init__(self._loss, array)
        self.loss_function = loss
        self._descriptors.pop("use_gpu", None)

        self.add_descriptors(loss=loss.__class__.__name__)

    def _loss(self, x: np.ndarray) -> float:
        image = self._generate_images(x)
        loss = self.loss_function(image)
        return loss

    def _generate_images(self, x: np.ndarray) -> np.ndarray:
        """Generates images tensor of shape [nb_images, x, y, 3] with pixels between 0 and 255"""
        # pylint: disable=not-callable
        noise = torch.tensor(x.astype("float32"))
        # type: ignore
        return (
            ((self.pgan_model.test(noise).clamp(min=-1, max=1) + 1) * 255.99 / 2)
            .permute(0, 2, 3, 1)
            .cpu()
            .numpy()
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import sys
import unittest
import platform
import subprocess
import typing as tp
from pathlib import Path
import pytest
import numpy as np
from . import testing


# decorators for tests which do not need to be run on Windows
skip_win = pytest.mark.skipif(sys.platform == "win32", reason="Internals")


@testing.parametrized(
    equal=([2, 3, 1], ""),
    missing=((1, 2), ["  - missing element(s): {3}."]),
    additional=((1, 4, 3, 2), ["  - additional element(s): {4}."]),
    both=((1, 2, 4), ["  - additional element(s): {4}.", "  - missing element(s): {3}."]),
)
def test_assert_set_equal(estimate: tp.Iterable[int], message: str) -> None:
    reference = {1, 2, 3}
    try:
        testing.assert_set_equal(estimate, reference)
    except AssertionError as error:
        if not message:
            raise AssertionError("An error has been raised while it should not.") from error
        np.testing.assert_equal(error.args[0].split("\n")[1:], message)
    else:
        if message:
            raise AssertionError("An error should have been raised.")


def test_printed_assert_equal() -> None:
    testing.printed_assert_equal(0, 0)
    np.testing.assert_raises(AssertionError, testing.printed_assert_equal, 0, 1)


@skip_win  # type: ignore
def test_assert_markdown_links_not_broken() -> None:
    folder = Path(__file__).parents[2].expanduser().absolute()
    assert (folder / "README.md").exists(), f"Wrong root folder: {folder}"
    assert testing._get_all_markdown_links(folder), "There should be at least one hyperlink!"
    testing.assert_markdown_links_not_broken(folder)


@testing.parametrized(
    changed=(RuntimeError, platform.system(), unittest.SkipTest),
    wrong_system=(RuntimeError, "blublu", RuntimeError),
    wrong_error=(AssertionError, platform.system(), RuntimeError),
)
def test_skip_test_on_system(
    skipped_error: tp.Type[Exception], system: str, expected_error: tp.Type[Exception]
) -> None:
    try:
        with pytest.raises(expected_error):
            with testing.skip_error_on_systems(skipped_error, (system,)):
                raise RuntimeError("Testing skip")
    except unittest.SkipTest as e:  # prevents SkipTest from just skipping the test and making it useless
        raise AssertionError("Should not have skipped the test!") from e


@skip_win  # type: ignore
def test_header() -> None:
    header = Path(__file__).read_text().splitlines()[0]
    repopath = Path(__file__).parents[1]
    assert repopath.name == "nevergrad"
    assert (repopath.parent / "setup.py").exists()
    output = subprocess.check_output(["find", str(repopath), "-name", "*.py"], shell=False).decode()
    missing: tp.List[str] = []
    for filepath in output.splitlines():
        if "lsgo" in filepath:
            continue
        if not Path(filepath).read_text().startswith(header):
            missing.append(filepath)
    if missing:
        missing_str = "\n - ".join(missing)
        raise AssertionError(
            f"Following files are missing standard header (see other files):\n - {missing_str}"
        )


@pytest.mark.skipif(sys.platform == "win32", reason="not compatible test")  # type: ignore
def test_no_import_warning() -> None:
    out = subprocess.run(
        ["python", "-c", "import nevergrad as ng"], check=True, stderr=subprocess.PIPE
    ).stderr
    assert not out



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from typing import List, Any, Callable
import numpy as np
from nevergrad.common import testing
from . import _funcs


@testing.parametrized(
    shifted_elliptic=(_funcs.ShiftedElliptic, ([12, 24],), [0, 10], 235571325.2279835),  # f1
    shifted_elliptic_0=(_funcs.ShiftedElliptic, ([12, 24],), [12, 24], 0),  # f1
    shifted_rastrigin=(_funcs.ShiftedRastrigin, ([12, 24],), [10, 20], 179.03902),  # f2
    shifted_rastrigin_0=(_funcs.ShiftedRastrigin, ([12, 24],), [12, 24], 0),  # f2
    shifted_ackley=(_funcs.ShiftedAckley, ([12, 24],), [10, 20], 18.1190856),  # f3
    shifted_ackley_0=(_funcs.ShiftedAckley, ([12, 24],), [12, 24], 0),  # f3
    shifted_rosenbrock=(_funcs.ShiftedRosenbrock, ([12, 24],), [10, 20], 6409.0),  # f12
    shifted_rosenbrock_0=(_funcs.ShiftedRosenbrock, ([12, 24],), [13, 25], 0),  # f12, shift + 1
    shifted_schwefel=(_funcs.ShiftedSchwefel, ([12, 24],), [10, 20], 40.48019),  # f15
    shifted_schwefel_0=(_funcs.ShiftedSchwefel, ([12, 24],), [12, 24], 0),  # f15
)
def test_testbed_functions(
    cls: Callable[..., Callable[[np.ndarray], float]], params: List[Any], data: List[float], expected: float
) -> None:
    np.random.seed(12)
    func = cls(*[np.array(p) for p in params])
    value = func(np.array(data))
    np.testing.assert_almost_equal(value, expected, decimal=5)


def test_read_data() -> None:
    data = _funcs.read_data("F4")
    testing.assert_set_equal(data.keys(), ["p", "s", "w", "xopt", "R50", "R25", "R100"])
    assert data["R50"].shape == (50, 50)
    np.testing.assert_almost_equal(
        data["R100"][:2, :2],
        [[-0.04915187244640615, 0.09617611414762256], [0.04609873606566875, 0.07060894383640659]],
    )


@testing.parametrized(  # manually verified against approximate Octave and CPP values on vector of zeros
    f1=(1, 1000, 209833896353.34344),
    f2=(2, 1000, 47620.31161660616),
    f3=(3, 1000, 21.72900253494246),  # Octave version needs uncommenting Tosz, Tasy, Lambda
    f4=(4, 1000, 107955147656065.94),
    f5=(5, 1000, 48419148.33292),
    f6=(6, 1000, 1077732.46530),  # Octave version needs uncommenting Tosz, Tasy, Lambda
    f7=(7, 1000, 993826981321072.6),  # same as implementations, which do not add transf on the side loss
    f8=(8, 1000, 5.722271501878064e18),
    f9=(9, 1000, 6001603202.501935),
    f10=(10, 1000, 98115481.6486656),  # Octave version uncommenting Tosz, Tasy, Lambda
    f11=(11, 1000, 1.0448520164721186e17),
    f12=(12, 1000, 1711354236949.7197),
    f13=(13, 905, 8.273800489859654e16),  # fails with octave
    f14=(14, 905, 4.4079796812096026e18),  # unspecified optimum
    f15=(15, 1000, 2393892336615502.5),
)
def test_functions_zeros(number: int, dimension: int, expected: float) -> None:
    func = _funcs.make_function(number)
    assert func.dimension == dimension
    value = func(np.array([0] * func.dimension))
    np.testing.assert_approx_equal(value, expected, significant=12)
    if number != 14:
        assert func.optimum is not None
        np.testing.assert_almost_equal(func(func.optimum), 0)


@testing.parametrized(  # verification against exact CPP outputs on vector of ones
    f1=(1, 15, 2.09946678145388153076e11),
    f2=(2, 15, 7.00495371043751511024e04),
    f3=(3, 15, 2.17108415925776405686e01),
    f4=(4, 15, 1.07162206769653859375e14),
    f5=(5, 15, 5.87148888268808051944e07),
    f6=(6, 15, 1.07977197180324327201e06),
    f7=(7, 14, 9.29113705518042875000e14),
    f8=(8, 15, 5.60788325599985049600e18),
    f9=(9, 15, 9.44072284529276657104e09),
    f10=(10, 13, 9.78947871248564124107e07),
    f11=(11, 15, 1.01442464039521104000e17),
    f12=(12, 15, 1.71217696529957031250e12),
    f13=(13, 15, 9.69220815693190400000e16),
    f14=(14, 14, 4.37551256977279180800e18),
    f15=(15, 15, 2.75152052424948050000e15),
)
def test_functions_ones(number: int, significant: int, expected: float) -> None:
    func = _funcs.make_function(number)
    value = func(np.array([1] * func.dimension))
    np.testing.assert_approx_equal(value, expected, significant=min(14, significant))



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
import nevergrad.common.typing as tp
from nevergrad.parametrization import parameter as p
from .hypervolume import HypervolumeIndicator


AUTO_BOUND = 15


# pylint: disable=too-many-instance-attributes
class HypervolumePareto:
    """Given several functions, and threshold on their values (above which solutions are pointless),
    this object can be used as a single-objective function, the minimization of which
    yields a solution to the original multiobjective problem.

    Parameters
    -----------
    upper_bounds: Tuple of float or np.ndarray
        upper_bounds[i] is a threshold above which x is pointless if function(x)[i] > upper_bounds[i].
    auto_bound: int
        if no upper bounds are provided, number of initial points used to estimate the upper bounds. Their
        loss will be 0 (except if they are uniformly worse than the previous points).
    seed: optional int or RandomState
        seed to use for selecting random subsamples of the pareto
    no_hypervolume: bool
        Most optimizers are designed for single objective and use a float loss.
        To use these in a multi-objective optimization, we provide the negative of
        the hypervolume of the pareto front as the loss.
        If not needed, an optimizer can set this to True.

    Notes
    -----
    When no_hypervolume is false:
    - This function is not stationary!
    - The minimum value obtained for this objective function is -h,
      where h is the hypervolume of the Pareto front obtained, given upper_bounds as a reference point.
    - The callable keeps track of the pareto_front (see attribute paretor_front) and is therefor stateful.
      For this reason it cannot be distributed. A user can however call the multiobjective_function
      remotely, and aggregate locally. This is what happens in the "minimize" method of optimizers.
    when no_hypervolume is true:
    - Hypervolume isn't used at all!
    - We simply add every point to the pareto front and state that the pareto front needs to be filtered.
    - The Pareto front is lazily kept up to date because every time you call pareto_front()
      an algorithm is performed that filters the pareto front into what it should be.
    """

    def __init__(
        self,
        *,
        upper_bounds: tp.Optional[tp.ArrayLike] = None,
        auto_bound: int = AUTO_BOUND,
        seed: tp.Optional[tp.Union[int, np.random.RandomState]] = None,
        no_hypervolume: bool = False,
    ) -> None:
        self._auto_bound = 0
        self._upper_bounds = (
            np.array([-float("inf")]) if upper_bounds is None else np.array(upper_bounds, copy=False)
        )
        if upper_bounds is None:
            self._auto_bound = auto_bound
        # If we are yet to set the upper bounds, or yet to have an add since doing so, _best_volume is -inf.
        # If we have set the upper bounds and all tells have been worse than them, then _best_volume is a
        # negative number indicating the least amount we have ever been worse.
        # If we have ever beaten the upper bounds, the _best_volume is nonnegative. In particular, it is
        # the hypervolume of the current PF if we are using hypervolume, otherwise 0.
        self._best_volume = -float("Inf")
        self._hypervolume: tp.Optional[HypervolumeIndicator] = None
        self._pareto_needs_filtering = False
        self._no_hypervolume = no_hypervolume
        self._pf = ParetoFront(seed=seed, no_hypervolume=no_hypervolume)

    @property
    def num_objectives(self) -> int:
        return self._upper_bounds.size

    @property
    def best_volume(self) -> float:
        return self._best_volume

    def extend(self, parameters: tp.Sequence[p.Parameter]) -> float:
        output = 0.0
        for param in parameters:
            output = self.add(param)
        return output

    def add(self, parameter: p.Parameter) -> float:
        """
        when _no_hypervolume = False
            Given parameters and the multiobjective loss, this computes the hypervolume
            and update the state of the function with new points if it belongs to the pareto front.
        when _no_hypervolume = True
            Add every point to pareto front. Don't compute hypervolume. Return 0.0 since loss
            not looked at in this context.
        """
        # pylint: disable=too-many-return-statements
        if not isinstance(parameter, p.Parameter):
            raise TypeError(
                f"{self.__class__.__name__}.add should receive a ng.p.Parameter, but got: {parameter}."
            )
        losses = parameter.losses
        if not isinstance(losses, np.ndarray):
            raise TypeError(
                f"Parameter should have multivalue as losses, but parameter.losses={losses} ({type(losses)})."
            )
        if self._auto_bound > 0:
            self._auto_bound -= 1
            if (self._upper_bounds > -float("inf")).all() and (losses > self._upper_bounds).all():
                return float("inf")  # Avoid uniformly worst points
            self._upper_bounds = np.maximum(self._upper_bounds, losses)
            self._pf.add_to_pareto(parameter)
            return 0.0
        # get rid of points over the upper bounds
        if (losses - self._upper_bounds > 0).any():
            loss = -float(np.sum(np.maximum(0, losses - self._upper_bounds)))
            if loss > self._best_volume:
                self._best_volume = loss
            if self._best_volume < 0:
                self._pf.add_to_pareto(parameter)
            return -loss
        if self._no_hypervolume:
            self._pf.add_to_pareto(parameter)
            return 0.0
        return self._calc_hypervolume(parameter, losses)

    def _calc_hypervolume(self, parameter: p.Parameter, losses: np.ndarray) -> float:
        if self._hypervolume is None:
            self._hypervolume = HypervolumeIndicator(self._upper_bounds)
            self._pf._hypervolume = self._hypervolume
        new_volume = self._hypervolume.compute([pa.losses for pa in self._pf.get_raw()] + [losses])
        if new_volume > self._best_volume:
            # This point is good! Let us give him a great mono-fitness value.
            self._best_volume = new_volume
            self._pf.add_to_pareto(parameter)
            return -new_volume
        else:
            # This point is not on the front
            # First we prune.
            distance_to_pareto = float("Inf")
            for param in self._pf.get_front():
                stored_losses = param.losses
                # TODO the following is probably not good at all:
                # -> +inf if no point is strictly better (but lower if it is)
                if (stored_losses <= losses).all():
                    distance_to_pareto = min(distance_to_pareto, min(losses - stored_losses))
            assert distance_to_pareto >= 0
            return 0.0 if self._no_hypervolume else -new_volume + distance_to_pareto

    def pareto_front(
        self, size: tp.Optional[int] = None, subset: str = "random", subset_tentatives: int = 12
    ) -> tp.List[p.Parameter]:
        """Pareto front, as a list of Parameter. The losses can be accessed through
        parameter.losses

        Parameters
        ------------
        size:  int (optional)
            if provided, selects a subset of the full pareto front with the given maximum size
        subset: str
            method for selecting the subset ("random, "loss-covering", "EPS", "domain-covering", "hypervolume")
            EPS is the epsilon indicator described e.g.
                here: https://hal.archives-ouvertes.fr/hal-01159961v2/document
        subset_tentatives: int
            number of random tentatives for finding a better subset

        Returns
        --------
        list
            the list of Parameter of the pareto front
        """
        return self._pf.get_front(size, subset, subset_tentatives)

    def get_min_losses(self) -> tp.List[float]:
        return np.min([p.losses for p in self._pf.get_raw()], axis=0)


class ParetoFront:
    def __init__(
        self,
        *,
        seed: tp.Optional[tp.Union[int, np.random.RandomState]] = None,
        no_hypervolume: bool = False,
    ) -> None:
        self._pareto: tp.List[p.Parameter] = []
        self._pareto_needs_filtering = False
        self._no_hypervolume = no_hypervolume
        self._rng = seed if isinstance(seed, np.random.RandomState) else np.random.RandomState(seed)
        self._hypervolume: tp.Optional[HypervolumeIndicator] = None

    def add_to_pareto(self, parameter: p.Parameter) -> None:
        self._pareto.append(parameter)
        self._pareto_needs_filtering = True

    def _filter_pareto_front(self) -> None:
        """Filters the Pareto front"""
        new_pareto: tp.List[p.Parameter] = []
        for param in self._pareto:  # quadratic :(
            should_be_added = True
            for other in self._pareto:
                if (other.losses <= param.losses).all() and (other.losses < param.losses).any():
                    should_be_added = False
                    break
            if should_be_added:
                new_pareto.append(param)
        self._pareto = new_pareto
        self._pareto_needs_filtering = False

    def get_raw(self) -> tp.List[p.Parameter]:
        """Retrieve current values, which may not be a Pareto front, as they have not been filtered."""
        return self._pareto

    # pylint: disable=too-many-branches
    def get_front(
        self, size: tp.Optional[int] = None, subset: str = "random", subset_tentatives: int = 12
    ) -> tp.List[p.Parameter]:
        """Pareto front, as a list of Parameter. The losses can be accessed through
        parameter.losses

        Parameters
        ------------
        size:  int (optional)
            if provided, selects a subset of the full pareto front with the given maximum size
        subset: str
            method for selecting the subset ("random, "loss-covering", "EPS", "domain-covering", "hypervolume")
            EPS is the epsilon indicator described e.g.
                here: https://hal.archives-ouvertes.fr/hal-01159961v2/document
        subset_tentatives: int
            number of random tentatives for finding a better subset

        Returns
        --------
        list
            the list of Parameter of the pareto front
        """
        if self._pareto_needs_filtering:
            self._filter_pareto_front()
        if size is None or size >= len(self._pareto):  # No limit: we return the full set.
            return self._pareto
        if subset == "random":
            return self._rng.choice(self._pareto, size).tolist()  # type: ignore
        tentatives = [self._rng.choice(self._pareto, size).tolist() for _ in range(subset_tentatives)]  # type: ignore
        if self._hypervolume is None and subset == "hypervolume":
            raise RuntimeError("Hypervolume subsetting not supported as hypervolume not in use")
        hypervolume = self._hypervolume
        scores: tp.List[float] = []
        for tentative in tentatives:
            if subset == "hypervolume":
                scores += [-hypervolume.compute([pa.losses for pa in tentative])]  # type: ignore
            else:
                score: float = 0.0
                for v in self._pareto:
                    best_score = float("inf") if subset != "EPS" else 0.0
                    for pa in tentative:
                        if subset == "loss-covering":  # Equivalent to IGD.
                            best_score = min(best_score, np.linalg.norm(pa.losses - v.losses))  # type: ignore
                        elif subset == "EPS":  # Cone Epsilon-Dominance.
                            best_score = min(best_score, max(pa.losses - v.losses))
                        elif subset == "domain-covering":
                            best_score = min(  # type: ignore
                                best_score, np.linalg.norm(pa.get_standardized_data(reference=v))
                            )  # TODO verify
                        else:
                            raise ValueError(f'Unknown subset for Pareto-Set subsampling: "{subset}"')
                    score += best_score**2 if subset != "EPS" else max(score, best_score)
                scores += [score]
        return tentatives[scores.index(min(scores))]  # type: ignore



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
import numpy as np
import nevergrad.common.typing as tp
from . import discretization
from . import core
from . import container
from . import _datalayers
from .data import Array

# weird pylint issue on "Descriptors"
# pylint: disable=no-value-for-parameter


C = tp.TypeVar("C", bound="Choice")
T = tp.TypeVar("T", bound="TransitionChoice")


class BaseChoice(container.Container):
    def __init__(
        self,
        *,
        choices: tp.Union[int, tp.Iterable[tp.Any]],
        repetitions: tp.Optional[int] = None,
        **kwargs: tp.Any,
    ) -> None:
        assert repetitions is None or isinstance(repetitions, int)  # avoid silent issues
        self._repetitions = repetitions
        # unroll iterables (includig Tuple instances or integers)
        lchoices = list(choices if not isinstance(choices, int) else range(choices))
        if not lchoices:
            raise ValueError("{self._class__.__name__} received an empty list of options.")
        super().__init__(choices=container.Tuple(*lchoices), **kwargs)

    def __len__(self) -> int:
        """Number of choices"""
        return len(self.choices)

    def _get_parameters_str(self) -> str:
        params = sorted(
            (k, p.name)
            for k, p in self._content.items()
            if p.name != self._ignore_in_repr.get(k, "#ignoredrepr#")
        )
        return ",".join(f"{k}={n}" for k, n in params)

    @property
    def index(self) -> int:  # delayed choice
        """Index of the chosen option"""
        inds = self.indices.value
        assert inds.size == 1
        return int(inds[0])

    @property
    def indices(self) -> Array:
        """Array of indices of the chosen option"""
        return self["indices"]  # type: ignore

    @property
    def choices(self) -> container.Tuple:
        """The different options, as a Tuple Parameter"""
        return self["choices"]  # type: ignore

    def _layered_get_value(self) -> tp.Any:
        if self._repetitions is None:
            return core.as_parameter(self.choices[self.index]).value
        return tuple(core.as_parameter(self.choices[ind]).value for ind in self.indices.value)

    def _layered_set_value(self, value: tp.List[tp.Any]) -> None:
        """Must be adapted to each class
        This handles a list of values, not just one
        """  # TODO this is currenlty very messy, may need some improvement
        values = [value] if self._repetitions is None else value
        self._check_frozen()
        indices: np.ndarray = -1 * np.ones(len(values), dtype=int)
        # try to find where to put this
        for i, val in enumerate(values):
            for k, choice in enumerate(self.choices):
                try:
                    choice.value = val
                    indices[i] = k
                    break
                except Exception:  # pylint: disable=broad-except
                    pass
            if indices[i] == -1:
                raise ValueError(f"Could not figure out where to put value {value}")
        self.indices.value = indices

    def get_value_hash(self) -> tp.Hashable:
        hashes: tp.List[tp.Hashable] = []
        for ind in self.indices.value:
            c = self.choices[int(ind)]
            const = isinstance(c, core.Constant) or not isinstance(c, core.Parameter)
            hashes.append(int(ind) if const else (int(ind), c.get_value_hash()))
        return tuple(hashes) if len(hashes) > 1 else hashes[0]


class Choice(BaseChoice):
    """Unordered categorical parameter, randomly choosing one of the provided choice options as a value.
    The choices can be Parameters, in which case there value will be returned instead.
    The chosen parameter is drawn randomly from the softmax of weights which are
    updated during the optimization.

    Parameters
    ----------
    choices: list or int
        a list of possible values or Parameters for the variable (or an integer as a shortcut for range(num))
    repetitions: None or int
        set to an integer :code:`n` if you want :code:`n` similar choices sampled independently (each with its own distribution)
        This is equivalent to :code:`Tuple(*[Choice(options) for _ in range(n)])` but can be
        30x faster for large :code:`n`.
    deterministic: bool
        whether to always draw the most likely choice (hence avoiding the stochastic behavior, but loosing
        continuity)

    Note
    ----
    - Since the chosen value is drawn randomly, the use of this variable makes deterministic
      functions become stochastic, hence "adding noise"
    - the "mutate" method only mutates the weights and the chosen Parameter (if it is not constant),
      leaving others untouched

    Examples
    --------

    >>> print(Choice(["a", "b", "c", "e"]).value)
    "c"

    >>> print(Choice(["a", "b", "c", "e"], repetitions=3).value)
    ("b", "b", "c")
    """

    def __init__(
        self,
        choices: tp.Union[int, tp.Iterable[tp.Any]],
        repetitions: tp.Optional[int] = None,
        deterministic: bool = False,
    ) -> None:
        lchoices = list(choices if not isinstance(choices, int) else range(choices))
        rep = 1 if repetitions is None else repetitions
        indices = Array(shape=(rep, len(lchoices)), mutable_sigma=False)
        indices.add_layer(_datalayers.SoftmaxSampling(len(lchoices), deterministic=deterministic))
        super().__init__(
            choices=lchoices,
            repetitions=repetitions,
            indices=indices,
        )
        self._indices: tp.Optional[np.ndarray] = None

    def mutate(self) -> None:
        # force random_state sync
        self.random_state  # pylint: disable=pointless-statement
        self.indices.mutate()
        for ind in self.indices.value:
            self.choices[ind].mutate()


class TransitionChoice(BaseChoice):
    """Categorical parameter, choosing one of the provided choice options as a value, with continuous transitions.
    By default, this is ordered, and most algorithms except discrete OnePlusOne algorithms will consider it as ordered.
    The choices can be Parameters, in which case there value will be returned instead.
    The chosen parameter is drawn using transitions between current choice and the next/previous ones.

    Parameters
    ----------
    choices: list or int
        a list of possible values or Parameters for the variable (or an integer as a shortcut for range(num))
    transitions: np.ndarray or Array
        the transition weights. During transition, the direction (forward or backward will be drawn with
        equal probabilities), then the transitions weights are normalized through softmax, the 1st value gives
        the probability to remain in the same state, the second to move one step (backward or forward) and so on.
    ordered: bool
        if False, changes the default behavior to be unordered and sampled uniformly when setting the data to a
        normalized and centered Gaussian (used in DiscreteOnePlusOne only)

    Note
    ----
    - the "mutate" method only mutates the weights and the chosen Parameter (if it is not constant),
      leaving others untouched
    - in order to support export to standardized space, the index is encoded as a scalar. A normal distribution N(O,1)
      on this scalar yields a uniform choice of index. This may come to evolve for simplicity's sake.
    - currently, transitions are computed through softmax, this may evolve since this is somehow impractical
    """

    def __init__(
        self,
        choices: tp.Union[int, tp.Iterable[tp.Any]],
        transitions: tp.Union[tp.ArrayLike, Array] = (1.0, 1.0),
        repetitions: tp.Optional[int] = None,
        ordered: bool = True,
    ) -> None:
        choices = list(choices if not isinstance(choices, int) else range(choices))
        indices = Array(init=len(choices) / 2.0 * np.ones((repetitions if repetitions is not None else 1,)))
        indices.set_bounds(0, len(choices), method="gaussian")
        indices = indices - 0.5
        intcasting = _datalayers.Int(deterministic=True)
        intcasting.arity = len(choices)
        indices.add_layer(intcasting)
        super().__init__(
            choices=choices,
            repetitions=repetitions,
            indices=indices,
            transitions=transitions if isinstance(transitions, Array) else np.array(transitions, copy=False),
        )
        assert self.transitions.value.ndim == 1
        self._ref: tp.Optional["TransitionChoice"] = None
        if not ordered:
            # always 0, standardized space of a transition choice is centered and Gaussian
            self._ref = self.copy()

    def _internal_set_standardized_data(self: T, data: np.ndarray, reference: T) -> None:
        ref = reference if self._ref is None else self._ref
        super()._internal_set_standardized_data(data, ref)  # type: ignore
        # force to bin
        super()._layered_set_value(super()._layered_get_value())

    def _internal_get_standardized_data(self: T, reference: T) -> np.ndarray:
        ref = reference if self._ref is None else self._ref
        return super()._internal_get_standardized_data(ref)  # type: ignore

    @property
    def transitions(self) -> Array:
        return self["transitions"]  # type: ignore

    def mutate(self) -> None:
        if self._ref is not None:
            new = self.random_state.normal(size=self.indices.value.size)
            self._internal_set_standardized_data(new, self._ref)
            return
        # force random_state sync
        self.random_state  # pylint: disable=pointless-statement
        transitions = core.as_parameter(self.transitions)
        transitions.mutate()
        rep = 1 if self._repetitions is None else self._repetitions
        #
        enc = discretization.Encoder(np.ones((rep, 1)) * np.log(self["transitions"].value), self.random_state)
        moves = enc.encode()
        signs = self.random_state.choice([-1, 1], size=rep)
        new_index = np.clip(self.indices.value + signs * moves, 0, len(self) - 1)
        self.indices.value = new_index
        # mutate corresponding parameter
        indices = set(self.indices.value)
        for ind in indices:
            self.choices[ind].mutate()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import typing as tp
import numpy as np
from nevergrad.common import testing
from . import mlfunctionlib


@testing.parametrized(
    decision_tree_depth=(
        dict(regressor="decision_tree_depth", data_dimension=1, dataset="artificial"),
        dict(depth=3),
        0.00116792703,
    ),
    decision_tree_artificial=(
        dict(regressor="decision_tree", data_dimension=12, dataset="artificial"),
        dict(depth=3, criterion="friedman_mse", min_samples_split=0.001),
        0.19138713337,
    ),
    mlp_artificial=(
        dict(regressor="mlp", data_dimension=2, dataset="artificial"),
        dict(activation="relu", solver="adam", alpha=0.01, learning_rate="constant"),
        0.01809799097751899,
    ),
    choosing_regressor_perceptron=(
        dict(regressor="any", data_dimension=2, dataset="artificial"),
        dict(
            activation="relu",
            solver="adam",
            alpha=0.01,
            learning_rate="constant",
            depth=3,
            criterion="friedman_mse",
            min_samples_split=0.001,
            regressor="decision_tree",
        ),
        0.0046690423,
    ),
    choosing_regressor_mlp=(
        dict(regressor="any", data_dimension=2, dataset="artificial"),
        dict(
            activation="relu",
            solver="adam",
            alpha=0.01,
            learning_rate="constant",
            depth=3,
            criterion="mse",
            min_samples_split=0.001,
            regressor="mlp",
        ),
        0.005673646,
    ),
    #    decision_tree_boston=(
    #        dict(regressor="decision_tree", data_dimension=None, dataset="boston"),
    #        dict(depth=5, criterion="friedman_mse", min_samples_split=0.001),
    #        24.985272503,
    #    ),
    decision_tree_diabetes=(
        dict(regressor="decision_tree", data_dimension=None, dataset="diabetes"),
        dict(depth=5, criterion="friedman_mse", min_samples_split=0.001),
        5057.790249631,
    ),
    decision_tree_cos=(
        dict(regressor="decision_tree", data_dimension=3, dataset="artificialcos"),
        dict(depth=5, criterion="friedman_mse", min_samples_split=0.001),
        0.00040148525,
    ),
    decision_tree_square=(
        dict(regressor="decision_tree", data_dimension=3, dataset="artificialsquare"),
        dict(depth=5, criterion="friedman_mse", min_samples_split=0.001),
        0.0019345073,
    ),
)
def test_mltuning_values(
    cls_params: tp.Dict[str, tp.Any], func_params: tp.Dict[str, tp.Any], expected: float
) -> None:

    np.random.seed(12)
    func = mlfunctionlib.MLTuning(**cls_params)
    if any(params.get("regressor", "") == "mlp" for params in (func_params, cls_params)):
        # make it way faster
        func.num_data = 12
        func._cross_val_num = 2
    outputs = [func(**func_params) for _ in range(2)]
    assert outputs[0] == outputs[1]  # function is deterministic once initialized
    np.testing.assert_almost_equal(outputs[0], expected, decimal=8)
    func.evaluation_function(func.parametrization)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# Based on https://github.com/Foloso/MixSimulator/tree/nevergrad_experiment

import numpy as np
from nevergrad.parametrization import parameter as p
from .. import base


class OptimizeFish(base.ExperimentFunction):
    """
    Fishing simulator.

    Parameters
    ----------
    time: int
        number of days of the planning

    """

    def __init__(self, time: int = 365) -> None:
        init = 0.5 * np.ones(time)
        parameter = p.Array(init=init)
        parameter.set_bounds(0.0 * init, 2.0 * init)
        parameter.set_name("raw_encoding")
        super().__init__(_compute_total_fishing, parameter)


def _compute_total_fishing(list_number_fishermen: np.ndarray) -> float:
    """Lotka-Volterra equations.

    This computes the total fishing, given the fishing effort every day.
    The problem makes sense for abritrary number of days, so that this works for
    any length of the input. 365 means one year."""
    number_dogfishes = 0.2
    number_haddocks = 0.8
    total_fishing = 0.0

    for number_fishermen in list(list_number_fishermen):
        number_fishermen = max(0, number_fishermen)
        # Number of haddocks eaten_haddocks today.
        number_eaten_haddocks = number_dogfishes * number_haddocks * 4 / 30
        # Haddock growth.
        number_haddocks += number_haddocks * 2 / 30
        number_haddocks -= number_eaten_haddocks
        # Natural growth of dogfishes.
        number_dogfishes -= (1 / 30) * number_dogfishes
        number_dogfishes += (5 / 2) * number_eaten_haddocks
        # Number of captured dogfishes.
        capture = min((1 / 10) * number_fishermen, number_dogfishes)
        number_dogfishes -= capture
        assert number_dogfishes >= 0.0
        total_fishing += capture
        number_dogfishes = min(1.0, number_dogfishes)
        number_haddocks = min(1.0, number_haddocks)
    return -total_fishing



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from nevergrad.common import errors
import nevergrad.common.typing as tp


class MultiobjectiveFunction:
    """MultiobjectiveFunction is deprecated and is removed after v0.4.3 "
    because it is no more needed. You should just pass a multiobjective loss to "
    optimizer.tell.\nSee https://facebookresearch.github.io/nevergrad/"
    optimization.html#multiobjective-minimization-with-nevergrad\n",
    """

    def __init__(
        self,
        multiobjective_function: tp.Callable[..., tp.ArrayLike],
        upper_bounds: tp.Optional[tp.ArrayLike] = None,
    ) -> None:
        raise errors.NevergradDeprecationError(
            "MultiobjectiveFunction is deprecated and is removed after v0.4.3 "
            "because it is no more needed. You should just pass a multiobjective loss to "
            "optimizer.tell.\nSee https://facebookresearch.github.io/nevergrad/"
            "optimization.html#multiobjective-minimization-with-nevergrad\n",
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import pytest

# import numpy as np
import sys

# from scipy import stats
# import nevergrad as ng


# decorators to be used when testing on Windows is unecessary
# or cumbersome
skip_win_perf = pytest.mark.skipif(
    sys.platform == "win32", reason="Slow, and no need to test performance on all platforms"
)


@skip_win_perf  # type: ignore
def test_sa() -> None:
    pass


#    num_tests = 77
#    for o in []:  # ["DiscreteOnePlusOne"]: dirty temporary hack.  # type: ignore
#        values = []
#        valuesT = []
#        for _ in range(num_tests):
#            dim = 30
#            arity = 3
#            budget = 57
#            domain = ng.p.TransitionChoice(range(arity), ordered=False, repetitions=dim)
#            optimum = np.random.randint(arity, size=dim)
#
#            def of(x):
#                return np.abs(np.abs(sum(x) - sum(optimum)) - 7.0)
#
#            recom = ng.optimizers.registry[o](domain, budget).minimize(of).value
#            recomT = ng.optimizers.registry["SA" + o + "Exp09"](domain, budget).minimize(of).value
#            values += [of(recom)]
#            valuesT += [of(recomT)]
#        pval = stats.mannwhitneyu(valuesT, values, alternative="less").pvalue
#        assert pval < 0.2, f"Simulated Annealing {o} fails the test: pval = {pval}."



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import inspect
from pathlib import Path
import numbers
import numpy as np
import nevergrad.common.typing as tp
from nevergrad.common import errors
from nevergrad.common.errors import (  # pylint: disable=unused-import
    UnsupportedExperiment as UnsupportedExperiment,
)
from nevergrad.parametrization import parameter as p
from nevergrad.optimization import multiobjective as mobj

EF = tp.TypeVar("EF", bound="ExperimentFunction")
ME = tp.TypeVar("ME", bound="MultiExperiment")


def _reset_copy(obj: p.Parameter) -> p.Parameter:
    """Copy a parameter and resets its random state to obtain variability"""
    out = obj.copy()
    out._set_random_state(None)  # propagates None to sub-parameters
    return out


# pylint: disable=too-many-instance-attributes
class ExperimentFunction:
    """Combines a function and its parametrization for running experiments (see benchmark subpackage)

    Parameters
    ----------
    function: callable
        the callable to convert
    parametrization: Parameter
        the parametrization of the function
    Notes
    -----
    - you can redefine custom "evaluation_function" and "compute_pseudotime" for custom behaviors in experiments
    - the bool/int/str/float init arguments are added as descriptors for the experiment which will serve in
      definining test cases. You can add more through "add_descriptors".
    - Makes sure you the "copy()" methods works (provides a new copy of the function *and* its parametrization)
      if you subclass ExperimentFunction since it is intensively used in benchmarks.
      By default, this will create a new instance using the same init arguments as your current instance
      (they were recorded through "__new__"'s magic) and apply the additional descriptors you may have added,
      as well as propagate the new parametrization *if it has a different name as the current one*.
    """

    def __new__(cls: tp.Type[EF], *args: tp.Any, **kwargs: tp.Any) -> EF:
        """Identifies initialization parameters during initialization and store them"""
        inst = object.__new__(cls)
        sig = inspect.signature(cls.__init__)
        callargs: tp.Dict[str, tp.Any] = {}
        try:
            boundargs = sig.bind(inst, *args, **kwargs)
        except TypeError:
            pass  # either a problem which will be caught later or a unpickling
        else:
            boundargs.apply_defaults()  # make sure we get the default non-provided arguments
            callargs = dict(boundargs.arguments)
            callargs.pop("self")
        inst._auto_init = callargs
        inst._descriptors = {
            x: y for x, y in callargs.items() if isinstance(y, (str, tuple, int, float, bool))
        }
        # if "bonnans" in str(cls.__name__) or "discrete" in str(cls.__name__) or "pbo" in str(cls.__name__):
        #    inst._descriptors = {
        #        x: y for x, y in callargs.items() if isinstance(y, (str, tuple, int, float, bool)) and "dimension" not
        #        in x and "paramet" not in x
        #    }
        inst._descriptors["function_class"] = cls.__name__
        return inst  # type: ignore

    def __init__(
        self: EF,
        function: tp.Callable[..., tp.Loss],
        parametrization: p.Parameter,
    ) -> None:
        assert callable(function)
        self._auto_init: tp.Dict[str, tp.Any]  # filled by __new__
        self._descriptors: tp.Dict[str, tp.Any]  # filled by __new__
        self._parametrization: p.Parameter
        self.parametrization = parametrization
        # force random state initialization
        self.multiobjective_upper_bounds: tp.Optional[np.ndarray] = None
        self.__function = function  # __ to prevent overrides
        # if this is not a function bound to this very instance, add the function/callable name to the descriptors
        if not hasattr(function, "__self__") or function.__self__ != self:  # type: ignore
            name = function.__name__ if hasattr(function, "__name__") else function.__class__.__name__
            self._descriptors.update(name=name)
        if len(self.parametrization.name) > 24:
            raise RuntimeError(
                f"For the sake of benchmarking, please rename the current parametrization:\n{self.parametrization!r}\n"
                "to a shorter name. This way it will be more readable in the experiments.\n"
                'Eg: parametrization.set_name("") to just ignore it\n'
                "CAUTION: Make sure you set different names for different parametrization configurations if you want it "
                "to be used in order to differentiate between benchmarks cases."
            )

    @property
    def dimension(self) -> int:
        return self._parametrization.dimension

    @property
    def parametrization(self) -> p.Parameter:
        return self._parametrization

    @parametrization.setter
    def parametrization(self, parametrization: p.Parameter) -> None:
        self._parametrization = parametrization
        self._parametrization.freeze()
        # pylint: disable=pointless-statement
        self._parametrization.random_state  # force initialization for synchronization of random state
        # # TODO investigate why this synchronization is needed

    @property
    def function(self) -> tp.Callable[..., tp.Loss]:
        return self.__function

    def __call__(self, *args: tp.Any, **kwargs: tp.Any) -> tp.Loss:
        """Call the function directly (equivalent to parametrized_function.function(*args, **kwargs))"""
        return self.function(*args, **kwargs)

    @property
    def descriptors(self) -> tp.Dict[str, tp.Any]:
        """Description of the function parameterization, as a dict. This base class implementation provides function_class,
        noise_level, transform and dimension
        """
        desc = dict(self._descriptors)  # Avoid external modification
        desc.update(parametrization=self.parametrization.name, dimension=self.dimension)
        return desc

    def add_descriptors(self, **kwargs: tp.Optional[tp.Hashable]) -> None:
        self._descriptors.update(kwargs)

    def __repr__(self) -> str:
        """Shows the function name and its summary"""
        params = [f"{x}={repr(y)}" for x, y in sorted(self._descriptors.items())]
        return "Instance of {}({})".format(self.__class__.__name__, ", ".join(params))

    def equivalent_to(self, other: tp.Any) -> bool:
        """Check that two instances where initialized with same settings.
        This is not meant to be used to check if functions are exactly equal
        (initialization may hold some randomness)
        This is only useful for unit testing.
        (may need to be overloaded to make faster if tests are getting slow)
        """
        if other.__class__ != self.__class__:
            return False
        return (
            bool(self._descriptors == other._descriptors)
            and self.parametrization.name == other.parametrization.name
        )

    def _internal_copy(self: EF) -> EF:
        """This is "black magic" which creates a new instance using the same init parameters
        that you provided and which were recorded through the __new__ method of ExperimentFunction
        """
        # auto_init is automatically filled by __new__, aka when creating the instance
        output: EF = self.__class__(
            **{x: _reset_copy(y) if isinstance(y, p.Parameter) else y for x, y in self._auto_init.items()}
        )
        return output

    def copy(self: EF) -> EF:
        """Provides a new equivalent instance of the class, possibly with
        different random initialization, to provide different equivalent test cases
        when using different seeds.
        This also checks that parametrization and descriptors are correct.
        You should preferably override _internal_copy
        """
        # add descriptors present in self but not added by initialization
        # (they must have been added manually)
        output = self._internal_copy()
        keys = set(output.descriptors)
        output.add_descriptors(**{x: y for x, y in self.descriptors.items() if x not in keys})
        # parametrization may have been overriden, so let's always update it
        # Caution: only if names differ!
        if output.parametrization.name != self.parametrization.name:
            output.parametrization = _reset_copy(self.parametrization)
        # then if there are still differences, something went wrong
        if not output.equivalent_to(self):
            raise errors.ExperimentFunctionCopyError(
                f"Copy of\n{self}\nwith descriptors:\n{self._descriptors}\nreturned non-equivalent\n"
                f"{output}\nwith descriptors\n{output._descriptors}.\n\n"
                "This means that the auto-copy behavior of ExperimentFunction does not work.\n"
                "You may want to implement your own copy method, or check implementation of "
                "ExperimentFunction.__new__ and copy to better understand what happens"
            )
        # propagate other useful information # TODO a bit hacky
        output.parametrization._constraint_checkers = self.parametrization._constraint_checkers
        output.multiobjective_upper_bounds = (
            self.multiobjective_upper_bounds
        )  # TODO not sure why this is needed
        return output

    def compute_pseudotime(  # pylint: disable=unused-argument
        self, input_parameter: tp.ArgsKwargs, loss: tp.Loss
    ) -> float:
        """Computes a pseudotime used during benchmarks for mocking parallelization in a reproducible way.
        By default, each call takes 1 unit of pseudotime, but this can be modified by overriding this
        function and the pseudo time can be a function of the function inputs and output.

        Note: This replaces get_postponing_delay which has been aggressively deprecated

        Parameters
        ----------
        input_parameter: Any
            the input that was provided to the actual function
        value: float
            the output of the actual function

        Returns
        -------
        float
            the pseudo computation time of the call to the actual function
        """
        return 1.0

    def evaluation_function(self, *recommendations: p.Parameter) -> float:
        """Provides the evaluation crieterion for the experiment.
        In case of mono-objective, it defers to evaluation_function
        Otherwise, it uses the hypervolume.
        This function can be overriden to provide custom behaviors.

        Parameters
        ----------
        *pareto: Parameter
            pareto front provided by the optimizer
        """

        if self.multiobjective_upper_bounds is None:  # singleobjective case
            assert len(recommendations) == 1
            output = self.function(*recommendations[0].args, **recommendations[0].kwargs)
            assert isinstance(
                output, numbers.Number
            ), f"evaluation_function can only be called on singleobjective experiments (output={output}) function={self.function}."
            return output  # type: ignore
        # multiobjective case
        hypervolume = mobj.HypervolumePareto(
            upper_bounds=self.multiobjective_upper_bounds, seed=self.parametrization.random_state
        )
        for candidate in recommendations:
            hypervolume.add(candidate)
        return -hypervolume.best_volume


def update_leaderboard(identifier: str, loss: float, array: np.ndarray, verbose: bool = True) -> None:
    """Handy function for storing best results for challenging functions (eg.: Photonics)
    The best results are kept in a file that is automatically updated with new data.
    This may require installing nevergrad in dev mode.

    Parameters
    ----------
    identifier: str
        the identifier of the problem
    loss: float
        the new loss, if better than the one in the file, the file will be updated
    array: np.ndarray
        the array corresponding to the loss
    verbose: bool
        whether to also print a message if the leaderboard was updated
    """
    # pylint: disable=import-outside-toplevel
    import pandas as pd  # lazzy to avoid requiring pandas for using an ExperimentFunction

    loss = np.round(loss, decimals=12)  # this is probably already too precise for the machine
    filepath = Path(__file__).with_name("leaderboard.csv")
    bests = pd.DataFrame(columns=["loss", "array"])
    if filepath.exists():
        bests = pd.read_csv(filepath, index_col=0)
    if identifier not in bests.index:
        bests.loc[identifier, :] = (float("inf"), "")
    try:
        if not bests.loc[identifier, "loss"] < loss:  # works for nan
            bests.loc[identifier, "loss"] = loss
            string = "[" + ",".join(str(x) for x in array.ravel()) + "]"
            bests.loc[identifier, "array"] = string
            bests = bests.loc[sorted(x for x in bests.index), :]
            bests.to_csv(filepath)
            if verbose:
                print(f"New best value for {identifier}: {loss}\nwith: {string[:80]}")
    except Exception:  # pylint: disable=broad-except
        pass  # better avoir bugs for this


class ArrayExperimentFunction(ExperimentFunction):
    """Combines a function and its parametrization for running experiments (see benchmark subpackage).
    Extends ExperimentFunction, in the special case of an array, by allowing the creation of symmetries
    of a single function. We can create ArrayExperimentFunction(callable, symmetry=i) for i in range(0, 2**d)
    when the callable works on R^d.
    Works only if there are no constraints.

    Parameters
    ----------
    function: callable
        the callable to convert
    parametrization: Parameter
        the parametrization of the function
    symmetry: int
        number parametrizing how we symmetrize the function.
    """

    def __init__(
        self, function: tp.Callable[..., tp.Loss], parametrization: p.Parameter, symmetry: int = 0
    ) -> None:
        """Adds a "symmetry" parameter, which allows the creation of many symmetries of a given function.

        symmetry: an int, 0 by default.
        if not zero, a symmetrization is applied to the input; each of the 2^d possible values
        for symmetry % 2^d gives one different function.
        Makes sense if and only if (1) the input is a single ndarray (2) the domains are symmetric."""
        self._inner_function = function
        super().__init__(self.symmetrized_function, parametrization)
        assert isinstance(
            parametrization, p.Array
        ), f"{type(parametrization)} is not p.Array; {parametrization}."
        assert (parametrization.bounds[0] is None) == (parametrization.bounds[1] is None)
        assert len(parametrization._constraint_checkers) == 0
        assert symmetry >= 0
        assert symmetry < 2**self.dimension
        # The number 11111111111111111111111 is prime (using a prime is an overkill but ok).
        symmetry = (symmetry * 11111111111111111111111) % (2**self.dimension)
        if symmetry != 0:
            self._function = self.symmetrized_function
            self.threshold_coefficients = np.zeros(self.dimension)
            self.slope_coefficients = np.ones(self.dimension)
            for i in range(self.dimension):  # pylint: disable=consider-using-enumerate
                if symmetry % 2 == 1:
                    if self.parametrization.bounds[0] is not None and self.parametrization.bounds[1] is not None:  # type: ignore
                        middle = (self.parametrization.bounds[0][0] + self.parametrization.bounds[1][0]) / 2.0  # type: ignore
                    else:
                        middle = 0.0
                    self.threshold_coefficients[i] = 2.0 * middle  # Otherwise we keep 0.
                    self.slope_coefficients[i] = -1.0  # Otherwise we keep 1.
                symmetry = symmetry // 2
        else:
            self._function = function
            self.threshold_coefficients = np.zeros(self.dimension)
            self.slope_coefficients = np.ones(self.dimension)

    def symmetrized_function(self, x: np.ndarray) -> tp.Loss:
        assert isinstance(x, np.ndarray), "symmetry != 0 works only when the input is an array."
        assert len(x.shape) == 1, "only one-dimensional arrays for now."
        return self._inner_function(self.threshold_coefficients + self.slope_coefficients * x)  # type: ignore


class MultiExperiment(ExperimentFunction):
    """Pack several mono-objective experiments into a multiobjective experiment


    Parameters
    ----------
    experiments: iterable of ExperimentFunction

    Notes
    -----
    - packing of multiobjective experiments is not supported.
    - parametrization must match between all functions (only their name is checked as initialization)
    - there is no descriptor for the packed functions, except the name (concatenetion of packed function names).
    """

    def __init__(
        self,
        experiments: tp.Iterable[ExperimentFunction],
        upper_bounds: tp.ArrayLike,
    ) -> None:
        xps = list(experiments)
        assert xps
        assert len(xps) == len({id(xp) for xp in xps}), "All experiments must be different instances"
        assert all(
            xp.multiobjective_upper_bounds is None for xp in xps
        ), "Packing multiobjective xps is not supported."
        assert all(
            xps[0].parametrization.name == xp.parametrization.name for xp in xps[1:]
        ), "Parametrization do not match"
        super().__init__(self._multi_func, xps[0].parametrization)
        self.multiobjective_upper_bounds = np.array(upper_bounds)
        self._descriptors.update(name=",".join(xp._descriptors.get("name", "#unknown#") for xp in xps))
        self._experiments = xps

    def _multi_func(self, *args: tp.Any, **kwargs: tp.Any) -> np.ndarray:
        outputs = [f(*args, **kwargs) for f in self._experiments]
        return np.array(outputs)

    def _internal_copy(self) -> "MultiExperiment":
        assert self.multiobjective_upper_bounds is not None
        return MultiExperiment([f.copy() for f in self._experiments], self.multiobjective_upper_bounds)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import functools
import math
import warnings
import weakref
import numpy as np
from scipy import optimize as scipyoptimize
import nevergrad.common.typing as tp
from nevergrad.parametrization import parameter as p
from nevergrad.common import errors
from . import base
from .base import IntOrParameter
from . import recaster


class _NonObjectMinimizeBase(recaster.SequentialRecastOptimizer):
    def __init__(
        self,
        parametrization: IntOrParameter,
        budget: tp.Optional[int] = None,
        num_workers: int = 1,
        *,
        method: str = "Nelder-Mead",
        random_restart: bool = False,
    ) -> None:
        super().__init__(parametrization, budget=budget, num_workers=num_workers)
        self.multirun = 1  # work in progress
        self._normalizer: tp.Any = None
        self.initial_guess: tp.Optional[tp.ArrayLike] = None
        # configuration
        # assert method in ["Nelder-Mead", "COBYLA", "SLSQP", "Powell", "BOBYQA", "AX"], f"Unknown method '{method}'"
        # assert method in ["SMAC3", "SMAC", "Nelder-Mead", "COBYLA", "SLSQP", "Powell"], f"Unknown method '{method}'"
        self.method = method
        self.random_restart = random_restart
        # self._normalizer = p.helpers.Normalizer(self.parametrization)
        assert (
            method
            in [
                "CmaFmin2",
                "gomea",
                "gomeablock",
                "gomeatree",
                "SMAC3",
                "BFGS",
                "LBFGSB",
                "L-BFGS-B",
                "SMAC",
                "AX",
                "Lamcts",
                "Nelder-Mead",
                "COBYLA",
                "BOBYQA",
                "SLSQP",
                "pysot",
                "negpysot",
                "Powell",
            ]
            or "NLOPT" in method
            or "BFGS" in method
        ), f"Unknown method '{method}'"
        if (
            method == "CmaFmin2"
            or "NLOPT" in method
            or "AX" in method
            or "BOBYQA" in method
            or "pysot" in method
            or "SMAC" in method
        ):
            normalizer = p.helpers.Normalizer(self.parametrization)
            #            if normalizer.fully_bounded or method == "AX" or "pysot" == method or "SMAC" in method:
            #                self._normalizer = normalizer
            self._normalizer = normalizer

    def _internal_tell_not_asked(self, candidate: p.Parameter, loss: tp.Loss) -> None:
        """Called whenever calling "tell" on a candidate that was not "asked".
        Defaults to the standard tell pipeline.
        """  # We do not do anything; this just updates the current best.

    def get_optimization_function(self) -> tp.Callable[[tp.Callable[[tp.ArrayLike], float]], tp.ArrayLike]:
        return functools.partial(self._optimization_function, weakref.proxy(self))

    @staticmethod
    def _optimization_function(
        weakself: tp.Any, objective_function: tp.Callable[[tp.ArrayLike], float]
    ) -> tp.ArrayLike:
        # pylint:disable=unused-argument
        budget = np.inf if weakself.budget is None else weakself.budget
        best_res = np.inf
        best_x: np.ndarray = weakself.current_bests["average"].x
        if weakself.initial_guess is not None:
            best_x = np.array(weakself.initial_guess, copy=True)  # copy, just to make sure it is not modified

        remaining: float = budget - weakself._num_ask

        def ax_obj(p):
            data = [p["x" + str(i)] for i in range(weakself.dimension)]  # type: ignore
            if weakself._normalizer:
                data = weakself._normalizer.backward(np.asarray(data, dtype=np.float_))
            return objective_function(data)

        while remaining > 0:  # try to restart if budget is not elapsed
            # print(f"Iteration with remaining={remaining}")
            options: tp.Dict[str, tp.Any] = {} if weakself.budget is None else {"maxiter": remaining}
            if weakself.method == "BOBYQA":
                import pybobyqa  # type: ignore

                res = pybobyqa.solve(objective_function, best_x, maxfun=budget, do_logging=False)
                if res.f < best_res:
                    best_res = res.f
                    best_x = res.x
            elif weakself.method == "AX":
                from ax import optimize as axoptimize  # type: ignore

                parameters = [
                    {"name": "x" + str(i), "type": "range", "bounds": [0.0, 1.0]}
                    for i in range(weakself.dimension)
                ]
                best_parameters, _best_values, _experiment, _model = axoptimize(
                    parameters, evaluation_function=ax_obj, minimize=True, total_trials=budget
                )
                best_x = np.array([float(best_parameters["x" + str(i)]) for i in range(weakself.dimension)])
                best_x = weakself._normalizer.backward(np.asarray(best_x, dtype=float))
            # options: tp.Dict[str, tp.Any] = {} if weakself.budget is None else {"maxiter": remaining}
            elif weakself.method[:5] == "NLOPT":
                # This is NLOPT, used as in the PCSE simulator notebook.
                # ( https://github.com/ajwdewit/pcse_notebooks ).
                import nlopt  # type: ignore

                def nlopt_objective_function(*args):
                    try:
                        data = np.asarray([arg for arg in args if len(arg) > 0])[0]
                    except Exception as e:
                        raise ValueError(f"{e}:\n{args}\n {[arg for arg in args]}")
                    assert len(data) == weakself.dimension, (
                        str(data) + " does not have length " + str(weakself.dimension)
                    )
                    if weakself._normalizer is not None:
                        data = weakself._normalizer.backward(np.asarray(data, dtype=np.float32))
                    return objective_function(data)

                # Sbplx (based on Subplex) is used by default.
                nlopt_param = (
                    getattr(nlopt, weakself.method[6:]) if len(weakself.method) > 5 else nlopt.LN_SBPLX
                )
                opt = nlopt.opt(nlopt_param, weakself.dimension)
                # Assign the objective function calculator
                opt.set_min_objective(nlopt_objective_function)
                # Set the bounds.
                opt.set_lower_bounds(np.zeros(weakself.dimension))
                opt.set_upper_bounds(np.ones(weakself.dimension))
                # opt.set_initial_step([0.05, 0.05])
                opt.set_maxeval(budget)

                # Start the optimization with the first guess
                firstguess = 0.5 * np.ones(weakself.dimension)
                best_x = opt.optimize(firstguess)
                # print("\noptimum at TDWI: %s, SPAN: %s" % (x[0], x[1]))
                # print("minimum value = ",  opt.last_optimum_value())
                # print("result code = ", opt.last_optimize_result())
                # print("With %i function calls" % objfunc_calculator.n_calls)
                if weakself._normalizer is not None:
                    best_x = weakself._normalizer.backward(np.asarray(best_x, dtype=np.float32))
            elif "pysot" in weakself.method:
                from poap.controller import BasicWorkerThread, ThreadController  # type: ignore

                from pySOT.experimental_design import SymmetricLatinHypercube  # type: ignore
                from pySOT.optimization_problems import OptimizationProblem  # type: ignore

                # from pySOT.strategy import SRBFStrategy
                from pySOT.strategy import DYCORSStrategy  # type: ignore
                from pySOT.surrogate import CubicKernel, LinearTail, RBFInterpolant  # type: ignore

                class LocalOptimizationProblem(OptimizationProblem):
                    def eval(self, data):
                        if weakself._normalizer is not None:
                            data = weakself._normalizer.backward(np.asarray(data, dtype=np.float32))
                        val = (
                            float(objective_function(data))
                            if "negpysot" not in weakself.method
                            else -float(objective_function(data))
                        )
                        return val

                dim = weakself.dimension
                opt_prob = LocalOptimizationProblem()
                opt_prob.dim = dim
                opt_prob.lb = np.array([0.0] * dim)
                opt_prob.ub = np.array([1.0] * dim)
                opt_prob.int_var = []
                opt_prob.cont_var = np.array(range(dim))

                rbf = RBFInterpolant(
                    dim=opt_prob.dim,
                    lb=opt_prob.lb,
                    ub=opt_prob.ub,
                    kernel=CubicKernel(),
                    tail=LinearTail(opt_prob.dim),
                )
                slhd = SymmetricLatinHypercube(dim=opt_prob.dim, num_pts=2 * (opt_prob.dim + 1))
                controller = ThreadController()
                # controller.strategy = SRBFStrategy(
                #    max_evals=budget, opt_prob=opt_prob, exp_design=slhd, surrogate=rbf, asynchronous=True
                # )
                controller.strategy = DYCORSStrategy(
                    opt_prob=opt_prob, exp_design=slhd, surrogate=rbf, max_evals=budget, asynchronous=True
                )
                worker = BasicWorkerThread(controller, opt_prob.eval)
                controller.launch_worker(worker)

                result = controller.run()

                best_res = result.value
                best_x = result.params[0]

            elif weakself.method == "SMAC3":

                # Import ConfigSpace and different types of parameters
                # from smac.configspace import ConfigurationSpace  # type: ignore  # noqa  # pylint: disable=unused-import
                # from smac.configspace import UniformFloatHyperparameter  # type: ignore
                # from smac.facade.smac_hpo_facade import SMAC4HPO  # type: ignore  # noqa  # pylint: disable=unused-import

                from ConfigSpace import Configuration, ConfigurationSpace, UniformFloatHyperparameter
                from smac import HyperparameterOptimizationFacade, Scenario

                # Import SMAC-utilities
                import threading
                import os
                import time
                from pathlib import Path

                the_date = str(time.time()) + "_" + str(np.random.rand())
                tag = str(np.random.rand())
                feed = "/tmp/smac_feed" + the_date + ".txt"
                fed = "/tmp/smac_fed" + the_date + ".txt"

                def dummy_function():
                    for _ in range(remaining):
                        # print(f"side thread waiting for request... ({u}/{weakself.budget})")
                        while (not Path(feed).is_file()) or os.stat(feed).st_size == 0:
                            time.sleep(0.1)
                        time.sleep(0.1)
                        # print("side thread happy to work on a request...")
                        data = np.loadtxt(feed)
                        os.remove(feed)
                        # print("side thread happy to really work on a request...")
                        res = objective_function(data)
                        # print("side thread happy to forward the result of a request...")
                        f = open(fed, "w")
                        f.write(str(res))
                        f.close()
                    return

                thread = threading.Thread(target=dummy_function)
                thread.start()

                # print(f"start SMAC3 optimization with budget {budget} in dimension {weakself.dimension}")
                cs = ConfigurationSpace()
                cs.add_hyperparameters(
                    [
                        UniformFloatHyperparameter(f"x{tag}{i}", 0.0, 1.0, default_value=0.0)
                        for i in range(weakself.dimension)
                    ]
                )

                def smac2_obj(p, seed: int = 0):
                    # print(f"SMAC3 proposes {p} {type(p)}")
                    pdata = [p[f"x{tag}{i}"] for i in range(len(p.keys()))]
                    data = weakself._normalizer.backward(np.asarray(pdata, dtype=float))
                    # print(f"converted to {data}")
                    if Path(fed).is_file():
                        os.remove(fed)
                    np.savetxt(feed, data)
                    while (not Path(fed).is_file()) or os.stat(fed).st_size == 0:
                        time.sleep(0.1)
                    time.sleep(0.1)
                    f = open(fed, "r")
                    res = float(f.read())
                    f.close()
                    # print(f"SMAC3 will receive {res}")
                    return res

                # scenario = Scenario({'cs': cs, 'run_obj': smac2_obj, 'runcount-limit': remaining, 'deterministic': True})
                scenario = Scenario(cs, deterministic=True, n_trials=int(remaining))

                smac = HyperparameterOptimizationFacade(scenario, smac2_obj)
                res = smac.optimize()
                best_x = np.array([res[f"x{tag}{k}"] for k in range(len(res.keys()))])
                best_x = weakself._normalizer.backward(np.asarray(best_x, dtype=float))
                # print(f"end SMAC optimization {best_x}")
                thread.join()
                weakself._num_ask = budget

            #            elif weakself.method == "SMAC":
            #                import smac  # noqa  # pylint: disable=unused-import
            #                import scipy.optimize  # noqa  # pylint: disable=unused-import
            #                from smac.facade.func_facade import fmin_smac  # noqa  # pylint: disable=unused-import
            #
            #                import threading
            #                import os
            #                import time
            #                from pathlib import Path
            #
            #                the_date = str(time.time())
            #                feed = "/tmp/smac_feed" + the_date + ".txt"
            #                fed = "/tmp/smac_fed" + the_date + ".txt"
            #
            #                def dummy_function():
            #                    for u in range(remaining):
            #                        print(f"side thread waiting for request... ({u}/{weakself.budget})")
            #                        while (not Path(feed).is_file()) or os.stat(feed).st_size == 0:
            #                            time.sleep(0.1)
            #                        time.sleep(0.1)
            #                        print("side thread happy to work on a request...")
            #                        data = np.loadtxt(feed)
            #                        os.remove(feed)
            #                        print("side thread happy to really work on a request...")
            #                        res = objective_function(data)
            #                        print("side thread happy to forward the result of a request...")
            #                        f = open(fed, "w")
            #                        f.write(str(res))
            #                        f.close()
            #                    return
            #
            #                thread = threading.Thread(target=dummy_function)
            #                thread.start()
            #
            #                def smac_obj(p):
            #                    print(f"SMAC proposes {p}")
            #                    data = weakself._normalizer.backward(
            #                        np.asarray([p[i] for i in range(len(p))], dtype=np.float)
            #                    )
            #                    print(f"converted to {data}")
            #                    if Path(fed).is_file():
            #                        os.remove(fed)
            #                    np.savetxt(feed, data)
            #                    while (not Path(fed).is_file()) or os.stat(fed).st_size == 0:
            #                        time.sleep(0.1)
            #                    time.sleep(0.1)
            #                    f = open(fed, "r")
            #                    res = np.float(f.read())
            #                    f.close()
            #                    print(f"SMAC will receive {res}")
            #                    return res
            #
            #                print(f"start SMAC optimization with budget {budget} in dimension {weakself.dimension}")
            #                assert budget is not None
            #                x, cost, _ = fmin_smac(
            #                    # func=lambda x: sum([(x_ - 1.234)**2  for x_ in x]),
            #                    func=smac_obj,
            #                    x0=[0.0] * weakself.dimension,
            #                    bounds=[(0.0, 1.0)] * weakself.dimension,
            #                    maxfun=remaining,
            #                    rng=weakself._rng.randint(5000),
            #                )  # Passing a seed makes fmin_smac determistic
            #                print("end SMAC optimization")
            #                thread.join()
            #                weakself._num_ask = budget
            #
            #                if cost < best_res:
            #                    best_res = cost
            #                    best_x = weakself._normalizer.backward(np.asarray(x, dtype=float))
            #

            elif "gomea" in weakself.method:
                import gomea

                class gomea_function(gomea.fitness.BBOFitnessFunctionRealValued):
                    def objective_function(self, objective_index, data):  # type: ignore
                        if weakself._normalizer is not None:
                            data = weakself._normalizer.backward(np.asarray(data, dtype=np.float32))
                        return objective_function(data)

                gomea_f = gomea_function(weakself.dimension)
                lm = {
                    "gomea": gomea.linkage.Univariate(),
                    "gomeablock": gomea.linkage.BlockMarginalProduct(2),
                    "gomeatree": gomea.linkage.LinkageTree("NMI".encode(), True, 0),
                }[weakself.method]
                rvgom = gomea.RealValuedGOMEA(
                    fitness=gomea_f,
                    linkage_model=lm,
                    lower_init_range=0.0,
                    upper_init_range=1.0,
                    max_number_of_evaluations=budget,
                )
                rvgom.run()
                best_x = gomea_f.best_x

            elif weakself.method == "CmaFmin2":
                import cma  # type: ignore

                def cma_objective_function(data):
                    # Hopefully the line below does nothing if unbounded and rescales from [0, 1] if bounded.
                    if weakself._normalizer is not None and weakself._normalizer.fully_bounded:
                        data = weakself._normalizer.backward(np.asarray(data, dtype=np.float32))
                    return objective_function(data)

                # cma.fmin2(objective_function, [0.0] * self.dimension, [1.0] * self.dimension, remaining)
                x0 = (
                    0.5 * np.ones(weakself.dimension)
                    if weakself._normalizer is not None and weakself._normalizer.fully_bounded
                    else np.zeros(weakself.dimension)
                )
                num_calls = 0
                while budget - num_calls > 0:
                    options = {"maxfevals": budget - num_calls, "verbose": -9}
                    if weakself._normalizer is not None and weakself._normalizer.fully_bounded:
                        # Tell CMA to work in [0, 1].
                        options["bounds"] = [0.0, 1.0]
                    res = cma.fmin(
                        cma_objective_function,
                        x0=x0,
                        sigma0=0.2,
                        options=options,
                        restarts=9,
                    )
                    x0 = (
                        0.5
                        + np.random.uniform() * np.random.uniform(low=-0.5, high=0.5, size=weakself.dimension)
                        if weakself._normalizer is not None and weakself._normalizer.fully_bounded
                        else np.random.randn(weakself.dimension)
                    )
                    if res[1] < best_res:
                        best_res = res[1]
                        best_x = res[0]
                        if weakself._normalizer is not None:
                            best_x = weakself._normalizer.backward(np.asarray(best_x, dtype=np.float32))
                    num_calls += res[2]
            else:
                res = scipyoptimize.minimize(
                    objective_function,
                    best_x
                    if not weakself.random_restart
                    else weakself._rng.normal(0.0, 1.0, weakself.dimension),
                    method=weakself.method,
                    options=options,
                    tol=0,
                )
                if res.fun < best_res:
                    best_res = res.fun
                    best_x = res.x
            remaining = budget - weakself._num_ask
        assert best_x is not None
        return best_x


class NonObjectOptimizer(base.ConfiguredOptimizer):
    """Wrapper over Scipy optimizer implementations, in standard ask and tell format.
    This is actually an import from scipy-optimize, including Sequential Quadratic Programming,

    Parameters
    ----------
    method: str
        Name of the method to use among:

        - Nelder-Mead
        - COBYLA
        - SQP (or SLSQP): very powerful e.g. in continuous noisy optimization. It is based on
          approximating the objective function by quadratic models.
        - Powell
        - NLOPT* (https://nlopt.readthedocs.io/en/latest/; by default, uses Sbplx, based on Subplex);
            can be NLOPT,
                NLOPT_LN_SBPLX,
                NLOPT_LN_PRAXIS,
                NLOPT_GN_DIRECT,
                NLOPT_GN_DIRECT_L,
                NLOPT_GN_CRS2_LM,
                NLOPT_GN_AGS,
                NLOPT_GN_ISRES,
                NLOPT_GN_ESCH,
                NLOPT_LN_COBYLA,
                NLOPT_LN_BOBYQA,
                NLOPT_LN_NEWUOA_BOUND,
                NLOPT_LN_NELDERMEAD.
    random_restart: bool
        whether to restart at a random point if the optimizer converged but the budget is not entirely
        spent yet (otherwise, restarts from best point)

    Note
    ----
    These optimizers do not support asking several candidates in a row
    """

    recast = True
    no_parallelization = True

    # pylint: disable=unused-argument
    def __init__(self, *, method: str = "Nelder-Mead", random_restart: bool = False) -> None:
        super().__init__(_NonObjectMinimizeBase, locals())


AX = NonObjectOptimizer(method="AX").set_name("AX", register=True)
BOBYQA = NonObjectOptimizer(method="BOBYQA").set_name("BOBYQA", register=True)
NelderMead = NonObjectOptimizer(method="Nelder-Mead").set_name("NelderMead", register=True)
CmaFmin2 = NonObjectOptimizer(method="CmaFmin2").set_name("CmaFmin2", register=True)
GOMEA = NonObjectOptimizer(method="gomea").set_name("GOMEA", register=True)
GOMEABlock = NonObjectOptimizer(method="gomeablock").set_name("GOMEABlock", register=True)
GOMEATree = NonObjectOptimizer(method="gomeatree").set_name("GOMEATree", register=True)
# NLOPT = NonObjectOptimizer(method="NLOPT").set_name("NLOPT", register=True)
Powell = NonObjectOptimizer(method="Powell").set_name("Powell", register=True)
RPowell = NonObjectOptimizer(method="Powell", random_restart=True).set_name("RPowell", register=True)
BFGS = NonObjectOptimizer(method="BFGS", random_restart=True).set_name("BFGS", register=True)
LBFGSB = NonObjectOptimizer(method="L-BFGS-B", random_restart=True).set_name("LBFGSB", register=True)
Cobyla = NonObjectOptimizer(method="COBYLA").set_name("Cobyla", register=True)
RCobyla = NonObjectOptimizer(method="COBYLA", random_restart=True).set_name("RCobyla", register=True)
SQP = NonObjectOptimizer(method="SLSQP").set_name("SQP", register=True)
SLSQP = SQP  # Just so that people who are familiar with SLSQP naming are not lost.
RSQP = NonObjectOptimizer(method="SLSQP", random_restart=True).set_name("RSQP", register=True)
RSLSQP = RSQP  # Just so that people who are familiar with SLSQP naming are not lost.
# NEWUOA = NonObjectOptimizer(method="NLOPT_LN_NEWUOA_BOUND").set_name("NEWUOA", register=True)
NLOPT_LN_SBPLX = NonObjectOptimizer(method="NLOPT_LN_SBPLX").set_name("NLOPT_LN_SBPLX", register=True)
NLOPT_LN_PRAXIS = NonObjectOptimizer(method="NLOPT_LN_PRAXIS").set_name("NLOPT_LN_PRAXIS", register=True)
NLOPT_GN_DIRECT = NonObjectOptimizer(method="NLOPT_GN_DIRECT").set_name("NLOPT_GN_DIRECT", register=True)
NLOPT_GN_DIRECT_L = NonObjectOptimizer(method="NLOPT_GN_DIRECT_L").set_name(
    "NLOPT_GN_DIRECT_L", register=True
)
NLOPT_GN_CRS2_LM = NonObjectOptimizer(method="NLOPT_GN_CRS2_LM").set_name("NLOPT_GN_CRS2_LM", register=True)
NLOPT_GN_AGS = NonObjectOptimizer(method="NLOPT_GN_AGS").set_name("NLOPT_GN_AGS", register=True)
NLOPT_GN_ISRES = NonObjectOptimizer(method="NLOPT_GN_ISRES").set_name("NLOPT_GN_ISRES", register=True)
NLOPT_GN_ESCH = NonObjectOptimizer(method="NLOPT_GN_ESCH").set_name("NLOPT_GN_ESCH", register=True)
NLOPT_LN_COBYLA = NonObjectOptimizer(method="NLOPT_LN_COBYLA").set_name("NLOPT_LN_COBYLA", register=True)
NLOPT_LN_BOBYQA = NonObjectOptimizer(method="NLOPT_LN_BOBYQA").set_name("NLOPT_LN_BOBYQA", register=True)
NLOPT_LN_NEWUOA_BOUND = NonObjectOptimizer(method="NLOPT_LN_NEWUOA_BOUND").set_name(
    "NLOPT_LN_NEWUOA_BOUND", register=True
)
NLOPT_LN_NELDERMEAD = NonObjectOptimizer(method="NLOPT_LN_NELDERMEAD").set_name(
    "NLOPT_LN_NELDERMEAD", register=True
)
# AX = NonObjectOptimizer(method="AX").set_name("AX", register=True)
# BOBYQA = NonObjectOptimizer(method="BOBYQA").set_name("BOBYQA", register=True)
# SMAC = NonObjectOptimizer(method="SMAC").set_name("SMAC", register=True)
SMAC3 = NonObjectOptimizer(method="SMAC3").set_name("SMAC3", register=True)


class _PymooMinimizeBase(recaster.SequentialRecastOptimizer):
    def __init__(
        self,
        parametrization: IntOrParameter,
        budget: tp.Optional[int] = None,
        num_workers: int = 1,
        *,
        algorithm: str,
    ) -> None:
        super().__init__(parametrization, budget=budget, num_workers=num_workers)
        # configuration
        self.algorithm = algorithm
        self._no_hypervolume = True
        self._initial_seed = -1

    def get_optimization_function(self) -> tp.Callable[[tp.Callable[..., tp.Any]], tp.Optional[tp.ArrayLike]]:
        if self._initial_seed == -1:
            self._initial_seed = self._rng.randint(2**30)
        return functools.partial(self._optimization_function, weakref.proxy(self))
        # pylint:disable=useless-return

    @staticmethod
    def _optimization_function(
        weakself: tp.Any, objective_function: tp.Callable[[tp.ArrayLike], float]
    ) -> tp.Optional[tp.ArrayLike]:
        # pylint:disable=unused-argument, import-outside-toplevel
        from pymoo import optimize as pymoooptimize

        from pymoo.factory import get_algorithm as get_pymoo_algorithm

        # from pymoo.factory import get_reference_directions

        # reference direction code for when we want to use the other MOO optimizers in Pymoo
        # if self.algorithm in [
        #     "rnsga2",
        #     "nsga3",
        #     "unsga3",
        #     "rnsga3",
        #     "moead",
        #     "ctaea",
        # ]:  # algorithms that require reference points or reference directions
        #     the appropriate n_partitions must be looked into
        #     ref_dirs = get_reference_directions("das-dennis", self.num_objectives, n_partitions=12)
        #     algorithm = get_pymoo_algorithm(self.algorithm, ref_dirs)
        # else:
        problem = _create_pymoo_problem(weakself, objective_function)
        if weakself.algorithm == "CMAES":
            from pymoo.algorithms.soo.nonconvex.cmaes import CMAES

            algorithm = CMAES(x0=np.random.random(problem.n_var), maxfevals=weakself.budget)
        elif weakself.algorithm == "BIPOP":
            from pymoo.algorithms.soo.nonconvex.cmaes import CMAES

            algorithm = CMAES(
                x0=np.random.random(problem.n_var),
                sigma=0.5,
                restarts=2,
                maxfevals=weakself.budget,
                tolfun=1e-6,
                tolx=1e-6,
                restart_from_best=True,
                bipop=True,
            )
        else:
            algorithm = get_pymoo_algorithm(weakself.algorithm)
        pymoooptimize.minimize(problem, algorithm, seed=weakself._initial_seed)
        return None

    def _internal_ask_candidate(self) -> p.Parameter:
        """
        Special version to make sure that num_objectives has been set before
        the proper _internal_ask_candidate, in our parent class, is called.
        """
        if self.num_objectives == 0:
            # dummy ask i.e. not activating pymoo until num_objectives is set
            warnings.warn(
                "with this optimizer, it is more efficient to set num_objectives before the optimization begins",
                errors.NevergradRuntimeWarning,
            )
            # We need to get a datapoint that is a random point in parameter space,
            # and waste an evaluation on it.
            return self.parametrization.spawn_child()
        return super()._internal_ask_candidate()

    def _internal_tell_candidate(self, candidate: p.Parameter, loss: float) -> None:
        """
        Special version to make sure that we the extra initial evaluation which
        we may have done in order to get num_objectives, is discarded.
        Note that this discarding means that the extra point will not make it into
        replay_archive_tell. Correspondingly, because num_objectives will make it into
        the pickle, __setstate__ will never need a dummy ask.
        """
        if self._messaging_thread is None:
            return  # dummy tell i.e. not activating pymoo until num_objectives is set
        super()._internal_tell_candidate(candidate, loss)

    def _post_loss(self, candidate: p.Parameter, loss: float) -> tp.Loss:
        # pylint: disable=unused-argument
        """
        Multi-Objective override for this function.
        """
        return candidate.losses


class Pymoo(base.ConfiguredOptimizer):
    """Wrapper over Pymoo optimizer implementations, in standard ask and tell format.
    This is actually an import from Pymoo Optimize.

    Parameters
    ----------
    algorithm: str

        Use "algorithm-name" with following names to access algorithm classes:
        Single-Objective
        -"de"
        -'ga'
        -"brkga"
        -"nelder-mead"
        -"pattern-search"
        -"cmaes"
        Multi-Objective
        -"nsga2"
        Multi-Objective requiring reference directions, points or lines
        -"rnsga2"
        -"nsga3"
        -"unsga3"
        -"rnsga3"
        -"moead"
        -"ctaea"

    Note
    ----
    These optimizers do not support asking several candidates in a row
    """

    recast = True
    no_parallelization = True

    # pylint: disable=unused-argument
    def __init__(self, *, algorithm: str) -> None:
        super().__init__(_PymooMinimizeBase, locals())


class _PymooBatchMinimizeBase(recaster.BatchRecastOptimizer):

    # pylint: disable=abstract-method

    def __init__(
        self,
        parametrization: IntOrParameter,
        budget: tp.Optional[int] = None,
        num_workers: int = 1,
        *,
        algorithm: str,
    ) -> None:
        super().__init__(parametrization, budget=budget, num_workers=num_workers)
        # configuration
        self.algorithm = algorithm
        self._no_hypervolume = True
        self._initial_seed = -1

    def get_optimization_function(self) -> tp.Callable[[tp.Callable[..., tp.Any]], tp.Optional[tp.ArrayLike]]:
        if self._initial_seed == -1:
            self._initial_seed = self._rng.randint(2**30)
        return functools.partial(self._optimization_function, weakref.proxy(self))
        # pylint:disable=useless-return

    @staticmethod
    def _optimization_function(
        weakself: tp.Any, objective_function: tp.Callable[[tp.ArrayLike], float]
    ) -> tp.Optional[tp.ArrayLike]:
        # pylint:disable=unused-argument, import-outside-toplevel
        from pymoo import optimize as pymoooptimize

        from pymoo.factory import get_algorithm as get_pymoo_algorithm

        # from pymoo.factory import get_reference_directions

        # reference direction code for when we want to use the other MOO optimizers in Pymoo
        # if self.algorithm in [
        #     "rnsga2",
        #     "nsga3",
        #     "unsga3",
        #     "rnsga3",
        #     "moead",
        #     "ctaea",
        # ]:  # algorithms that require reference points or reference directions
        #     the appropriate n_partitions must be looked into
        #     ref_dirs = get_reference_directions("das-dennis", self.num_objectives, n_partitions=12)
        #     algorithm = get_pymoo_algorithm(self.algorithm, ref_dirs)
        # else:
        algorithm = get_pymoo_algorithm(weakself.algorithm)
        problem = _create_pymoo_problem(weakself, objective_function, False)
        pymoooptimize.minimize(problem, algorithm, seed=weakself._initial_seed)
        return None

    def _internal_ask_candidate(self) -> p.Parameter:
        """Reads messages from the thread in which the underlying optimization function is running
        New messages are sent as "ask".
        """
        # get a datapoint that is a random point in parameter space
        if self.num_objectives == 0:  # dummy ask i.e. not activating pymoo until num_objectives is set
            warnings.warn(
                "with this optimizer, it is more efficient to set num_objectives before the optimization begins",
                errors.NevergradRuntimeWarning,
            )
            return self.parametrization.spawn_child()
        return super()._internal_ask_candidate()

    def _internal_tell_candidate(self, candidate: p.Parameter, loss: float) -> None:
        """Returns value for a point which was "asked"
        (none asked point cannot be "tell")
        """
        if self._messaging_thread is None:
            return  # dummy tell i.e. not activating pymoo until num_objectives is set
        super()._internal_tell_candidate(candidate, loss)

    def _post_loss(self, candidate: p.Parameter, loss: float) -> tp.Loss:
        # pylint: disable=unused-argument
        """
        Multi-Objective override for this function.
        """
        return candidate.losses


class PymooBatch(base.ConfiguredOptimizer):
    """Wrapper over Pymoo optimizer implementations, in standard ask and tell format.
    This is actually an import from Pymoo Optimize.

    Parameters
    ----------
    algorithm: str

        Use "algorithm-name" with following names to access algorithm classes:
        Single-Objective
        -"de"
        -'ga'
        -"brkga"
        -"nelder-mead"
        -"pattern-search"
        -"cmaes"
        Multi-Objective
        -"nsga2"
        Multi-Objective requiring reference directions, points or lines
        -"rnsga2"
        -"nsga3"
        -"unsga3"
        -"rnsga3"
        -"moead"
        -"ctaea"

    Note
    ----
    These optimizers do not support asking several candidates in a row
    """

    recast = True

    # pylint: disable=unused-argument
    def __init__(self, *, algorithm: str) -> None:
        super().__init__(_PymooBatchMinimizeBase, locals())


def _create_pymoo_problem(
    optimizer: base.Optimizer,
    objective_function: tp.Callable[[tp.ArrayLike], float],
    elementwise: bool = True,
):
    kwargs = {}
    try:
        # pylint:disable=import-outside-toplevel
        from pymoo.core.problem import ElementwiseProblem, Problem  # type: ignore

        Base = ElementwiseProblem if elementwise else Problem
    except ImportError:
        # Used if pymoo < 0.5.0
        # pylint:disable=import-outside-toplevel
        from pymoo.model.problem import Problem as Base  # type: ignore

        kwargs = {"elementwise_evaluation": elementwise}

    class _PymooProblem(Base):  # type: ignore
        def __init__(self, optimizer, objective_function):
            self.objective_function = objective_function
            super().__init__(
                n_var=optimizer.dimension,
                n_obj=optimizer.num_objectives,
                n_constr=0,  # constraints handled already by nevergrad
                xl=-math.pi * 0.5,
                xu=math.pi * 0.5,
                **kwargs,
            )

        def _evaluate(self, X, out, *args, **kwargs):
            # pylint:disable=unused-argument
            # pymoo is supplying us with bounded parameters in [-pi/2,pi/2]. Nevergrad wants unbounded reals from us.
            out["F"] = self.objective_function(np.tan(X))

    return _PymooProblem(optimizer, objective_function)


PymooCMAES = Pymoo(algorithm="CMAES").set_name("PymooCMAES", register=True)
PymooBIPOP = Pymoo(algorithm="BIPOP").set_name("PymooBIPOP", register=True)
PymooNSGA2 = Pymoo(algorithm="nsga2").set_name("PymooNSGA2", register=True)


##Not yet included, coming.
# from .lamcts.MCTS import lamcts_minimize
#
#
# class _LamctsMinimizeBase(recaster.SequentialRecastOptimizer):
#    def __init__(
#        self,
#        parametrization: IntOrParameter,
#        budget: tp.Optional[int] = None,
#        num_workers: int = 1,
#        *,
#        method: str = "Nelder-Mead",
#        random_restart: bool = False,
#        device: str = "cpu",
#    ) -> None:
#        super().__init__(parametrization, budget=budget, num_workers=num_workers)
#        self.multirun = 1  # work in progress
#        normalizer = p.helpers.Normalizer(self.parametrization)
#        self._normalizer = normalizer
#
#        self.initial_guess: tp.Optional[tp.ArrayLike] = None
#        # configuration
#        assert method in ["Nelder-Mead", "COBYLA", "SLSQP", "Powell"], f"Unknown method '{method}'"
#        self.method = method
#        self.random_restart = random_restart
#        self.device = device
#
#    def _internal_tell_not_asked(self, candidate: p.Parameter, value: float) -> None:
#        """Called whenever calling "tell" on a candidate that was not "asked".
#        Defaults to the standard tell pipeline.
#        """  # We do not do anything; this just updates the current best.
#
#    def get_optimization_function(self) -> tp.Callable[[tp.Callable[[tp.ArrayLike], float]], tp.ArrayLike]:
#        # create a different sub-instance, so that the current instance is not referenced by the thread
#        # (consequence: do not create a thread at initialization, or we get a thread explosion)
#        subinstance = self.__class__(
#            parametrization=self.parametrization,
#            budget=self.budget,
#            num_workers=self.num_workers,
#            method=self.method,
#            random_restart=self.random_restart,
#        )
#        subinstance.archive = self.archive
#        subinstance.current_bests = self.current_bests
#        return subinstance._optimization_function
#
#    def _optimization_function(self, objective_function: tp.Callable[[tp.ArrayLike], float]) -> tp.ArrayLike:
#        # pylint:disable=unused-argument
#        budget = np.inf if self.budget is None else self.budget
#        best_res = np.inf
#        best_x: np.ndarray = self.current_bests["average"].x  # np.zeros(self.dimension)
#        if self.initial_guess is not None:
#            best_x = np.array(self.initial_guess, copy=True)  # copy, just to make sure it is not modified
#        remaining = budget - self._num_ask
#        while remaining > 0:  # try to restart if budget is not elapsed
#            options: Dict[str, int] = {} if self.budget is None else {"maxiter": remaining}
#
#            def lamcts_obj(data):
#                # print("transform", data)
#                data = (data + 1.0) / 2.0
#                data = self._normalizer.backward(np.asarray(data, dtype=np.float))
#                return objective_function(data)
#
#            res = lamcts_minimize(
#                # func=objective_function,
#                func=lamcts_obj,
#                dims=self.parametrization.dimension,
#                budget=self.budget,
#                device=self.device,
#                #                best_x if not self.random_restart else self._rng.normal(0.0, 1.0, self.dimension),
#                #                method=self.method,
#                #                options=options,
#                #                tol=0,
#            )
#            # def lamcts_minimize(func, dims, budget, lb=None, ub=None):
#            if res.fun < best_res:
#                best_res = res.fun
#                best_x = res.x
#                best_x = 2.0 * weakself._normalizer.backward(np.asarray(best_x, dtype=np.float32)) - 1.0
#            remaining = budget - self._num_ask
#        return best_x
#
#
# class LamctsOptimizer(base.ConfiguredOptimizer):
#    """Wrapper over Lamcts optimizer implementations, in standard ask and tell format.
#    Sequential Quadratic Programming. Inside Nevergrad, this code is in https://github.com/facebookresearch/nevergrad/blob/master/nevergrad/optimization/optimizerlib.py; this is actually an import from scipy-optimize. It is very powerful e.g. in continuous noisy optimization. It is based on approximating the objective function by quadratic models.
#
#        Parameters
#        ----------
#        method: str
#            Name of the method to use among:
#
#            - Nelder-Mead
#            - COBYLA
#            - SQP (or SLSQP): very powerful e.g. in continuous noisy optimization. It is based on
#              approximating the objective function by quadratic models.
#            - Powell
#        random_restart: bool
#            whether to restart at a random point if the optimizer converged but the budget is not entirely
#            spent yet (otherwise, restarts from best point)
#
#        Note
#        ----
#        These optimizers do not support asking several candidates in a row
#    """
#
#    recast = True
#    no_parallelization = True
#
#    # pylint: disable=unused-argument
#    def __init__(self, *, random_restart: bool = False, device: str = "cpu") -> None:
#        super().__init__(_LamctsMinimizeBase, locals())
#

PymooBatchNSGA2 = PymooBatch(algorithm="nsga2").set_name("PymooBatchNSGA2", register=False)
pysot = NonObjectOptimizer(method="pysot").set_name("pysot", register=True)
# negpysot = NonObjectOptimizer(method="negpysot").set_name("negpysot", register=True)
